import numpy as np
import imageio
import argparse
import sys
import random

DEBUG = False

TIME = 1.2 #1.75
P_DENSITY = 1
FRICTION = 0.05
DEPOSITION_RATE=0.15
P_EVAPORATION=0.001
P_SCALE = 1
P_MINVOL = 0.01
DEFAULT_NUM_ITERATIONS = 10000

parser = argparse.ArgumentParser()
parser.add_argument(
    '--input',
    type=str,
    required=True
)
parser.add_argument(
    '--iterations',
    type=int,
    default=DEFAULT_NUM_ITERATIONS
)
args = parser.parse_args()

map = imageio.imread(args.input)
WIDTH = map.shape[0]
HEIGHT = map.shape[1]
NUM_MOVES_PER_DROPLET = WIDTH*2

height = np.array([[map[x][y][0] for y in range(HEIGHT)] for x in range(WIDTH)], dtype=np.float32)

class Particle:
    """
    Based off of: https://nickmcd.me/2020/04/10/simple-particle-based-hydraulic-erosion/
    """
    def __init__(self):
        # Total particle volume
        self.volume = 1
        # Amount of sediment that we are carrying
        self.sediment = 0

        self.x = random.randint(1, WIDTH-2)
        self.y = random.randint(1, HEIGHT-2)
        # Speed
        self.sx = 0
        self.sy = 0

    def _surface_normal(self, i, j):
        """
        See: https://github.com/weigert/SimpleErosion/blob/master/source/include/world/world.cpp
        """
        def normalize_then_multiply(x, y, z, c):
            mag = np.sqrt(x**2 + y**2 + z**2)
            return (c*x/mag, c*y/mag, c*z/mag)

        def add(A, B):
            return (A[0] + B[0], A[1] + B[1], A[2] + B[2])
        
        n = (0, 0, 0)
        n = add(n, normalize_then_multiply(P_SCALE*(height[i][j]-height[i+1][j]), 1, 0, 0.15))
        n = add(n, normalize_then_multiply(P_SCALE*(height[i-1][j]-height[i][j]), 1, 0, 0.15))
        n = add(n, normalize_then_multiply(0, 1, P_SCALE*(height[i][j]-height[i+1][j+1]), 0.15))
        n = add(n, normalize_then_multiply(0, 1, P_SCALE*(height[i][j-1]-height[i+1][j]), 0.15))
        n = add(n, normalize_then_multiply(P_SCALE*(height[i][j] - height[i+1][j+1])/np.sqrt(2), np.sqrt(2), P_SCALE*(height[i][j] - height[i+1][j+1])/np.sqrt(2), 0.1))
        n = add(n, normalize_then_multiply(P_SCALE*(height[i][j] - height[i+1][j-1])/np.sqrt(2), np.sqrt(2), P_SCALE*(height[i][j] - height[i+1][j-1])/np.sqrt(2), 0.1))
        n = add(n, normalize_then_multiply(P_SCALE*(height[i][j] - height[i-1][j+1])/np.sqrt(2), np.sqrt(2), P_SCALE*(height[i][j] - height[i-1][j+1])/np.sqrt(2), 0.1))
        n = add(n, normalize_then_multiply(P_SCALE*(height[i][j] - height[i-1][j-1])/np.sqrt(2), np.sqrt(2), P_SCALE*(height[i][j] - height[i-1][j-1])/np.sqrt(2), 0.1))

        return n

    def _out_of_bounds(self, x, y):
        return x < 1 or y < 1 or x >= WIDTH-1 or y >= HEIGHT-1


    def erode(self):
        self.lifetime = 0
        while self.volume > P_MINVOL:
            self.lifetime += 1
            old_loc = [int(self.x), int(self.y)]

            if self._out_of_bounds(self.x, self.y):
                break
            
            N = self._surface_normal(int(self.x), int(self.y))

            self.sx += TIME*N[0]/(self.volume*P_DENSITY)
            self.sy += TIME*N[2]/(self.volume*P_DENSITY)
            self.x += TIME*self.sx
            self.y += TIME*self.sy
            self.sx *= (1-TIME*FRICTION)
            self.sy *= (1-TIME*FRICTION)

            if self._out_of_bounds(self.x, self.y):
                break

            # Equilibrium sediment content
            old_height = height[old_loc[0]][old_loc[1]]
            new_height = height[int(self.x)][int(self.y)]
            max_sediment = self.volume * np.sqrt(self.sx**2 + self.sy**2) * (new_height - old_height)
            max_sediment = max(max_sediment, 0)

            # Capacity difference
            sdiff = max_sediment - self.sediment

            # Mass transfer
            self.sediment += TIME*DEPOSITION_RATE*sdiff
            height[int(self.x)][int(self.y)] -= TIME*self.volume*DEPOSITION_RATE*sdiff

            # Volume transfer
            self.volume *= (1 - TIME*P_EVAPORATION)

total_lifetime = 0
for i in range(args.iterations):
    percent = int(args.iterations / 100)
    if not DEBUG:
        if i % percent == 0:
            print('.', end='')
            sys.stdout.flush()
    particle = Particle()
    particle.erode()
    total_lifetime += particle.lifetime

print()
print('Summary:')
print('Average lifespan:', int(total_lifetime / args.iterations), 'ticks')
# print('Average time each cell was traversed:', round_num(np.mean(density), precision=2), 'times')


# Re-write b&w map. Construct it as a 32 bit array even though we don't really
# need to.
out = np.empty((WIDTH, HEIGHT, 4))
for x in range(WIDTH):
    for y in range(HEIGHT):
        out[x][y] = [height[x][y], height[x][y], height[x][y], 255]
path = args.input.split('.')
imageio.imwrite(path[0] + '-out.' + path[1], out.astype(np.uint8))