import imageio
import random
import numpy as np
import math
import sys
import argparse
import sklearn.preprocessing

#
# Global variables
#
DEBUG = False
VERBOSE_DEBUG = DEBUG and True
DEFAULT_NUM_ITERATIONS = 10000

# Default parameters. These work well, but not on the scale of Pavala...
# P_INERTIA = 0.025
# P_MIN_SLOPE = 0.05
# P_CAPACITY = 1
# P_DEPOSITION = 0.02
# P_EROSION = 0.1
# P_GRAVITY = 20
# P_EVAPORATION = 0.001

P_INERTIA = 0.1
P_MIN_SLOPE = 0.05
P_CAPACITY = 1
P_DEPOSITION = 0.02
P_EROSION = 0.04
P_GRAVITY = 20
P_EVAPORATION = 0.001

parser = argparse.ArgumentParser()
parser.add_argument(
    '--input',
    type=str,
    required=True
)
parser.add_argument(
    '--iterations',
    type=int,
    default=DEFAULT_NUM_ITERATIONS
)
args = parser.parse_args()

map = imageio.imread(args.input)
WIDTH = map.shape[0]
HEIGHT = map.shape[1]
NUM_MOVES_PER_DROPLET = WIDTH*2

height = np.array([[map[x][y][0] for y in range(HEIGHT)] for x in range(WIDTH)], dtype=np.float32)
density = np.zeros(np.shape(height))

# Stats
total_lifespan = 0

def round_num(n, precision=4):
    return np.format_float_positional(n, precision=precision, unique=False, fractional=False, trim='k')

def in_bounds(x, y):
    return x >= 0 and x < WIDTH and y >= 0 and y < HEIGHT

def on_map_edge(x, y):
    return x < 1 or x >= WIDTH-1 or y < 1 or y >= HEIGHT-1

def deposit(x, y, x_r, y_r, amt) -> float:
    """
    Deposit at (x, y) and surrounding tiles and then return the
    new height value of (x, y)
    """
    def deposit_at(x, y, amt):
        if not in_bounds(x, y):
            return
        height[x][y] += amt
    
    deposit_at(x, y, (1-x_r)*(1-y_r)*amt)
    deposit_at(x+1, y, x_r*(1-y_r)*amt)
    deposit_at(x, y+1, (1-x_r)*y_r*amt)
    deposit_at(x+1, y+1, x_r*y_r*amt)
    return height[x][y]

def erode(x, y, amt) -> float:
    """
    Erode at (x, y) and surrounding tiles and then return the
    new height value of (x, y)
    """
    # I don't understand it, I just copied it from ranmantaru's blog.
    FUNNY_CONSTANT = 0.1591549430918953

    def erode_at(x, y, amt):
        height[x][y] -= amt

    for i in range(x-1, x+3):
        dx = x - i
        for j in range(y-1, y+3):
            dy = y - j
            if not in_bounds(i, j):
                continue
            # The proportion of all sediment to drop here
            p = (1 - (dx ** 2 + dy ** 2) * 0.25) * FUNNY_CONSTANT
            if p <= 0:
                continue
            erode_at(i, j, p*amt)

    return height[x][y]

def simulate_droplet(plot_trajectory=False):
    """
    Adapted from
    https://www.firespark.de/resources/downloads/implementation%20of%20a%20methode%20for%20hydraulic%20erosion.pdf
    http://ranmantaru.com/blog/2011/10/08/water-erosion-on-heightmap-terrain/
    """
    start_x = random.randint(1, WIDTH-2)
    start_y = random.randint(1, HEIGHT-2)
    trajectory = []

    # Initialize the important variables
    x = start_x
    x_int = start_x
    x_r = 0
    y = start_y
    y_int = start_y
    y_r = 0
    sediment = 0
    velocity = 0
    water = 1
    # (dx, dy): the direction we are moving in, maintained across iterations to provide 'inertia'
    dx = 0 
    dy = 0

    # The heights of this point and its neighbors, used to calculate the gradient.
    h = height[x][y]
    h00 = h
    h10 = height[x+1][y]
    h01 = height[x][y+1]
    h11 = height[x+1][y+1]

    for i in range(0, NUM_MOVES_PER_DROPLET):
        density[x_int][y_int] += velocity

        if plot_trajectory:
            trajectory.append([x_int, y_int])

        # Use gradient to find next position
        if VERBOSE_DEBUG:
            print(f'x={round_num(x)}\t\ty={round_num(y)}\t\tv={round_num(velocity)}')
        gx = h00 + h01 - h10 - h11
        gy = h00 + h10 - h01 - h11
        dx = (dx-gx)*P_INERTIA+gx
        dy = (dy-gy)*P_INERTIA+gy
        magnitude = np.sqrt(dx**2 + dy**2)
        
        if magnitude < 1:
            # Pick a random direction to move in
            theta = random.random() * 2 * math.pi
            dx = np.cos(theta)
            dy = np.sin(theta)
        else:
            # Normalize
            dx /= magnitude
            dy /= magnitude

        if VERBOSE_DEBUG:
            print(f'gx={round_num(gx)}\tgy={round_num(gx)}\tdy={round_num(dy)}\tdx={round_num(dx)}')

        nx = x + dx
        ny = y + dy

        # If we're at the edge just give up and go off it I guess...?
        if on_map_edge(nx, ny):
            if VERBOSE_DEBUG:
                print('going off map edge')
            break
        
        # Sample the new height ('nh'). Use the remainders 'nx_r' and 'ny_r' that we rounded by
        # to interpolate.
        nx_int = int(nx)
        ny_int = int(ny)
        nx_r = nx - nx_int
        ny_r = ny - ny_int
        nh00 = height[nx_int][ny_int]
        nh10 = height[nx_int+1][ny_int]
        nh01 = height[nx_int][ny_int+1]
        nh11 = height[nx_int+1][ny_int+1]
        nh = (nh00*(1-nx_r)+nh10*nx_r)*(1-ny_r)+(nh01*(1-nx_r)+nh11*nx_r)*ny_r

        if VERBOSE_DEBUG:
            print(f'h={round_num(h)}\t\tnh={round_num(nh)}\t{("(going uphill)" if nh > h else "")}')

        if nh > h:
            # We should try to deposit sediment to match the new height.
            ds = nh - h
            if ds >= sediment:
                # We don't have enough sediment so let's just dump it all and
                # call it a day. This droplet's work is done.
                ds = sediment
                h = deposit(x_int, y_int, x_r, y_r, ds)
                sediment = 0
                if VERBOSE_DEBUG:
                    print('ran out of sediment, dying')
                break
            else:
                # We have more than enough sediment, so the droplet will live on... for now.
                # Deposit enough to match the new height and then kill our velocity so we
                # don't keep going uphill.
                h = deposit(x_int, y_int, x_r, y_r, ds)
                sediment -= ds
                velocity = 0
        
        # Compute the transport capacity. This is how much sediment we can carry
        # at this time and place. Then use that to figure out the change in sediment we carry (w.r.t ground)
        dh = h - nh
        capacity = max(dh, P_MIN_SLOPE) * velocity * water * P_CAPACITY + 2*np.sqrt(density[x_int][y_int] + 1)*P_CAPACITY
        ds = sediment - capacity


        if ds >= 0:
            # Deposit some amount of the sediment we are carrying into the new tile and adjacent.
            ds *= P_DEPOSITION
            if VERBOSE_DEBUG:
                print(f'cap.={round_num(capacity)}\tds={round_num(ds)}\t(depositing)')
            nh = deposit(nx_int, ny_int, nx_r, ny_r, ds)
            sediment -= ds
        else:
            # Pick up some amount of sediment from this area.
            ds *= -P_EROSION
            if VERBOSE_DEBUG:
                print(f'cap.={round_num(capacity)}\tds={round_num(ds)}\t(eroding)')
            nh = erode(x_int, y_int, ds)
            dh -= ds
            sediment += ds
        
        # Wrap up: assign variables for next tick
        velocity = (max(dh, 0) + velocity) / 2
        water *= 1 - P_EVAPORATION
        x = nx
        x_int = nx_int
        x_r = nx_r
        y = ny
        y_int = ny_int
        y_r = ny_r
        h = nh
        h00 = nh00
        h01 = nh01
        h10 = nh10
        h11 = nh11
        if VERBOSE_DEBUG:
            print()

    # Make sure we dump the remaining sediment that we have
    deposit(x_int, y_int, x_r, y_r, sediment)

    if DEBUG:
        print(f'stats: dist.={round_num(np.sqrt((start_x - x)**2 + (start_y - y)**2), 2)}\tlife={i}')
    if VERBOSE_DEBUG:
        print(f'---------------------------------------------------')

    # Keep track of stats
    global total_lifespan
    total_lifespan += i
    return trajectory

#
# Process then write to file
#
for x in range(args.iterations):
    percent = int(args.iterations / 100)
    if not DEBUG:
        if x % percent == 0:
            print('.', end='')
            sys.stdout.flush()
    simulate_droplet()

print()
print('Summary:')
print('Average lifespan:', int(total_lifespan / args.iterations), 'ticks')
print('Average time each cell was traversed:', round_num(np.mean(density), precision=2), 'times')

# Re-write b&w map. Construct it as a 32 bit array even though we don't really
# need to.
out = np.empty((WIDTH, HEIGHT, 4))
for x in range(WIDTH):
    for y in range(HEIGHT):
        out[x][y] = [height[x][y], height[x][y], height[x][y], 255]
path = args.input.split('.')
imageio.imwrite(path[0] + '-out.' + path[1], out.astype(np.uint8))

# Write density map, after normalizing.
density /= np.mean(density)
density *= 64
density = np.clip(density, 0, 255)
imageio.imwrite(path[0] + '-out-density.' + path[1], density.astype(np.uint8))

# Plot 5 sample trajectories
colors = [
    [0, 0, 255, 255],
    [0, 255, 0, 255],
    [0, 255, 255, 255],
    [255, 0, 255, 255],
    [255, 255, 0, 255]
]
trajectoryMap = np.copy(out)#np.full((WIDTH, HEIGHT, 4), [0, 0, 0, 255])
for color in colors:
    trajectory = simulate_droplet(plot_trajectory=True)
    for x, y in trajectory:
        trajectoryMap[x][y] = color
imageio.imwrite(path[0] + '-out-trajectories.' + path[1], trajectoryMap.astype(np.uint8))